import logging
from functools import partial, wraps

import numpy as np
import treefiles as tf
from gcma.wrapper import CMA


def d1(func):
    @wraps(func)
    def _func(obj, list_params, n_iter, **kwargs):
        pa = np.array([x.vals for x in list_params])
        res = func(obj, pa, **kwargs)[:, np.newaxis]
        res = np.repeat(res, len(obj.features), axis=1)
        fe = [obj.features.assign(x).set_error_L2() for x in res]
        return np.array([x.errors for x in fe]), res

    return _func


class CMATestFunctions(CMA):
    test_functions = ("sphere", "rastrigin", "bukin")

    def set_test(self, name, **kwargs):
        if hasattr(self, name):
            setattr(self, "energy", partial(getattr(self, name), **kwargs))
        else:
            log.error(f"Test function {name!r} not found")

    def energy(self, *args, **kwargs):
        log.error("energy: you must call `set_test` to attach a test function")
        raise NotImplementedError

    @d1
    def sphere(self, x):
        return np.einsum("ij,ij->i", x, x)

    @d1
    def rastrigin(self, x, a=10):
        return a * x.shape[1] + (np.square(x) - a * np.cos(2 * np.pi * x)).sum(axis=1)

    @d1
    def bukin(self, x):
        return (
            100 * np.sqrt(np.abs(x[:, 1] - 0.01 * np.square(x[:, 0])))
            + 0.01 * np.abs(x[:, 0] + 10)
            + 13
        )


search_space = {
    "sphere": [(-10, 10)],
    "rastrigin": [(-5.12, 5.12)],
    "bukin": [(-15, -5), (-3, 3)],
}


log = logging.getLogger(__name__)

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    log = tf.get_logger()
